# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from ctypes import *
from .ktx_error_code import KtxErrorCode, KtxError
from .ktx_hash_list import KtxHashList
from pyktx.native import ffi, lib
from typing import Dict, Literal, Optional


class KtxVersionMismatchError(Exception):
    """Error thrown when reading a file with wrong KTX version."""

    pass


class KtxTexture:
    """
    Base class representing a texture.

    ktxTextures should be created only by one of the static factory methods,
    and these fields should be considered read-only.
    """

    def __init__(self, ptr: c_uint64):
        self._ptr = ptr
        self.__kv_data_head = KtxHashList(lib.PY_ktxTexture_get_kvDataHead(ptr))

    def __del__(self):
        lib.free(self._ptr)
        self._ptr = ffi.NULL

    @property
    def class_id(self) -> int:
        """Identify the class type. 1 for KtxTexture1, 2 for KtxTexture2."""

        return lib.PY_ktxTexture_get_classId(self._ptr)

    @property
    def is_array(self) -> bool:
        """true if the texture is an array texture, i.e, a GL_TEXTURE_*_ARRAY target is to be used."""

        return lib.PY_ktxTexture_get_isArray(self._ptr)

    @property
    def is_compressed(self) -> bool:
        """If the texture's format is a block compressed format."""

        return lib.PY_ktxTexture_get_isCompressed(self._ptr)

    @property
    def is_cubemap(self) -> bool:
        """
        True if the texture is a cubemap or cubemap array.
        """
        return lib.PY_ktxTexture_get_isCubemap(self._ptr)

    @property
    def generate_mipmaps(self) -> bool:
        """If mipmaps should be generated for the texture when uploading to graphics APIs."""

        return lib.PY_ktxTexture_get_generateMipmaps(self._ptr)

    @property
    def base_width(self) -> int:
        """Width of the texture's base level."""

        return lib.PY_ktxTexture_get_baseWidth(self._ptr)

    @property
    def base_height(self) -> int:
        """Height of the texture's base level."""

        return lib.PY_ktxTexture_get_baseHeight(self._ptr)

    @property
    def base_depth(self) -> int:
        """Depth of the texture's base level."""

        return lib.PY_ktxTexture_get_baseDepth(self._ptr)

    @property
    def num_dimensions(self) -> int:
        """Number of dimensions in the texture: 1, 2 or 3."""

        return lib.PY_ktxTexture_get_numDimensions(self._ptr)

    @property
    def num_levels(self) -> int:
        """Number of mip levels in the texture."""

        return lib.PY_ktxTexture_get_numLevels(self._ptr)

    @property
    def num_faces(self) -> int:
        """Number of faces: 6 for cube maps, 1 otherwise."""

        return lib.PY_ktxTexture_get_numFaces(self._ptr)

    @property
    def element_size(self) -> int:
        """The element size of the texture's images."""

        return lib.ktxTexture_GetElementSize(self._ptr)

    @property
    def kv_data_raw(self) -> Optional[c_buffer]:
        """
        The raw KV data buffer.

        This is available only if RAW_KVDATA_BIT was used in create-flag bits.
        """

        buffer = lib.PY_ktxTexture_get_kvData(self._ptr)

        if buffer == ffi.NULL:
            return None

        return ffi.buffer(buffer, lib.PY_ktxTexture_get_kvDataLen(self._ptr))

    @property
    def kv_data(self) -> KtxHashList:
        """
        The metadata stored in the texture as a hash-list.

        This is not available if SKIP_KVDATA_BIT was used in the create-flag bits.
        """

        return self.__kv_data_head

    @property
    def data_size(self) -> int:
        """The total size of the texture image data in bytes."""

        return lib.ktxTexture_GetDataSize(self._ptr)

    @property
    def data_size_uncompressed(self) -> int:
        """Byte length of the texture's uncompressed image data."""

        return lib.ktxTexture_GetDataSizeUncompressed(self._ptr)

    def row_pitch(self, level: int) -> int:
        """
        Return pitch between rows of a texture image level in bytes.

        For uncompressed textures the pitch is the number of bytes between
        rows of texels. For compressed textures it is the number of bytes
        between rows of blocks. The value is padded to GL_UNPACK_ALIGNMENT,
        if necessary. For all currently known compressed formats padding will
        not be necessary.
        """

        return lib.ktxTexture_GetRowPitch(self._ptr, level)

    def image_size(self, level: int) -> int:
        """
        Calculate & return the size in bytes of an image at the specified mip level.

        For arrays, this is the size of layer, for cubemaps, the size of a face and
        for 3D textures, the size of a depth slice.

        The size reflects the padding of each row to KTX_GL_UNPACK_ALIGNMENT.
        """

        return lib.ktxTexture_GetImageSize(self._ptr, level)

    def image_offset(self, level: int, layer: int, face_slice: int) -> int:
        """
        Find the offset of an image within a ktxTexture's image data.

        As there is no such thing as a 3D cubemap we make the 3rd location parameter
        do double duty.
        """

        data = lib.PY_ktxTexture_GetImageOffset(self._ptr, level, layer, face_slice)

        if int(data.error) != KtxErrorCode.SUCCESS:
            raise KtxError('ktxTexture_GetImageOffset', KtxErrorCode(data.error))

        return data.offset

    def data(self) -> bytes:
        """Return a buffer holding the texture image data."""

        return ffi.buffer(lib.ktxTexture_GetData(self._ptr), self.data_size)

    def set_image_from_memory(self, level: int, layer: int, face_slice: int, data: bytes):
        """
        Set image for level, layer, faceSlice from an image in memory.

        Uncompressed images in memory are expected to have their rows tightly packed
        as is the norm for most image file formats. The copied image is padded as
        necessary to achieve the KTX-specified row alignment. No padding is done if the
        ktxTexture's is_compressed field is true. Level, layer, face_slice rather than offset
        are specified to enable some validation.
        """

        error = KtxErrorCode(lib.ktxTexture_SetImageFromMemory(self._ptr, level, layer, face_slice, data, len(data)))

        if int(error) != KtxErrorCode.SUCCESS:
            raise KtxError('ktxTexture_SetImageFromMemory', KtxErrorCode(int(error)))

    def write_to_named_file(self, dst_name: str) -> None:
        """Save this texture to a named file in KTX format."""

        error = KtxErrorCode(lib.ktxTexture_WriteToNamedFile(self._ptr, dst_name.encode('ascii')))

        if int(error) != KtxErrorCode.SUCCESS:
            raise KtxError('ktxTexture_WriteToNamedFile', KtxErrorCode(error))

    def write_to_native_memory(self) -> Dict[Literal["bytes", "size"], int]:
        """Write this KTX file to block of memory in KTX format. Recommended to use write_to_memory instead."""

        data = lib.PY_ktxTexture_WriteToMemory(self._ptr)

        if int(data.error) != KtxErrorCode.SUCCESS:
            raise KtxError('ktxTexture_WriteToMemory', KtxErrorCode(data.error))

        return {"bytes": data.bytes, "size": data.size}

    def write_to_memory(self) -> bytes:
        """Write this KTX file to a buffer in KTX format."""

        data = self.write_to_native_memory()
        native_buffer = ffi.buffer(data.get('bytes'), data.get('size'))
        buffer = native_buffer[:]

        lib.free(data.get('bytes'))
        return buffer



# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from pyktx import *
from test_config import __test_images__
import unittest


class TestKtxTexture2(unittest.TestCase):
    def test_create_from_named_file(self):
        test_ktx_file = os.path.join(__test_images__, 'astc_ldr_4x4_FlightHelmet_baseColor.ktx2')
        texture = KtxTexture2.create_from_named_file(test_ktx_file)

        self.assertEqual(texture.num_levels, 1)
        self.assertEqual(texture.num_faces, 1)
        self.assertEqual(texture.vk_format, VkFormat.VK_FORMAT_ASTC_4x4_SRGB_BLOCK)
        self.assertEqual(texture.base_width, 2048)
        self.assertEqual(texture.base_height, 2048)
        self.assertEqual(texture.supercompression_scheme, KtxSupercmpScheme.NONE)

    def test_create_from_named_file_mipmapped(self):
        test_ktx_file = os.path.join(__test_images__, 'astc_mipmap_ldr_4x4_posx.ktx2')
        texture = KtxTexture2.create_from_named_file(test_ktx_file, KtxTextureCreateFlagBits.NO_FLAGS)

        self.assertEqual(texture.num_levels, 12)
        self.assertEqual(texture.base_width, 2048)
        self.assertEqual(texture.base_height, 2048)

    def test_get_image_size(self):
        test_ktx_file = os.path.join(__test_images__, 'astc_mipmap_ldr_4x4_posx.ktx2')
        texture = KtxTexture2.create_from_named_file(test_ktx_file)

        self.assertEqual(texture.image_size(0), 4194304)

    def test_get_image_offset(self):
        test_ktx_file = os.path.join(__test_images__, 'astc_mipmap_ldr_4x4_posx.ktx2')
        texture = KtxTexture2.create_from_named_file(test_ktx_file)

        level11_offset = texture.image_offset(11, 0, 0)
        level0_offset = texture.image_offset(0, 0, 0)

        self.assertEqual(level11_offset, 0)
        self.assertEqual(level0_offset - level11_offset, 0x155790 - 0x220)

    def test_get_size(self):
        test_ktx_file = os.path.join(__test_images__, 'astc_mipmap_ldr_4x4_posx.ktx2')
        texture = KtxTexture2.create_from_named_file(test_ktx_file)

        self.assertEqual(texture.num_levels, 12)

        data_size = texture.data_size
        total_size = 0

        for i in range(0, 12):
            total_size += texture.image_size(i)

        self.assertEqual(total_size, data_size)

        data = texture.data()
        self.assertEqual(len(data), data_size)

    def test_get_data(self):
        test_ktx_file = os.path.join(__test_images__, 'astc_mipmap_ldr_4x4_posx.ktx2')
        texture = KtxTexture2.create_from_named_file(test_ktx_file)

        self.assertEqual(texture.num_levels, 12)

        level0_length = texture.image_size(0)

        with open(test_ktx_file, 'rb') as file:
            data = texture.data()
            self.assertEqual(file.read()[-level0_length - 1:], data[-level0_length - 1:])

    def test_compress_basis(self):
        test_ktx_file = os.path.join(__test_images__, 'arraytex_7_mipmap_reference_u.ktx2')
        texture = KtxTexture2.create_from_named_file(test_ktx_file, KtxTextureCreateFlagBits.LOAD_IMAGE_DATA_BIT)

        self.assertEqual(texture.is_compressed, False)
        self.assertEqual(texture.supercompression_scheme, KtxSupercmpScheme.NONE)

        texture.compress_basis(KtxBasisParams(quality_level=1))

        self.assertEqual(texture.is_compressed, True)
        self.assertEqual(texture.supercompression_scheme, KtxSupercmpScheme.BASIS_LZ)

    def test_transcode_basis(self):
        test_ktx_file = os.path.join(__test_images__, 'color_grid_basis.ktx2')
        texture = KtxTexture2.create_from_named_file(test_ktx_file, KtxTextureCreateFlagBits.LOAD_IMAGE_DATA_BIT)

        texture.transcode_basis(KtxTranscodeFmt.ASTC_4x4_RGBA, 0)

        self.assertEqual(texture.vk_format, VkFormat.VK_FORMAT_ASTC_4x4_SRGB_BLOCK)

    def test_create(self):
        info = KtxTextureCreateInfo(
            gl_internal_format=None,  # ignored
            base_width=10,
            base_height=10,
            base_depth=1,
            vk_format=VkFormat.VK_FORMAT_ASTC_4x4_SRGB_BLOCK)

        texture = KtxTexture2.create(info, KtxTextureCreateStorage.ALLOC)
        self.assertEqual(texture.vk_format, VkFormat.VK_FORMAT_ASTC_4x4_SRGB_BLOCK)
        texture.set_image_from_memory(0, 0, 0, bytes(texture.data_size))



# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from enum import IntEnum


class KtxPackAstcEncoderMode(IntEnum):
    """
    Options specifying ASTC encoder profile mode.

    This and function is used later to derive the profile.
    """

    DEFAULT = 0
    LDR = 1
    HDR = 2
    MAX = HDR



# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

import os
from setuptools import setup

__name__    = 'pyktx'
__version__ = os.environ['LIBKTX_VERSION']

assert __version__ is not None

setup(
    name=__name__,
    version=__version__,
    description='A Python interface to the libktx library',
    author='Shukant Pal',
    author_email='foss@shukantpal.com',
    cffi_modules=["buildscript.py:ffibuilder"],
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    include_package_data=True,
    install_requires=["cffi>=1.15.1"],
    license="Apache 2.0",
    license_files=('LICENSE'),
    long_description_content_type="text/markdown",
    long_description="This Python package provides a Pythonic interface to libktx. It uses CFFI to generate the C bindings.",
    packages=['pyktx'],
    package_dir={'pyktx': 'pyktx'},
    setup_requires=["cffi>=1.15.1"],
    url='https://github.com/KhronosGroup/KTX-Software',
)



# Copyright (c) 2023, Shukant Pal and Contributors
# SPDX-License-Identifier: Apache-2.0

from cffi import FFI
import platform
import sys
import os
import unittest

LIBKTX_INSTALL_DIR = os.getenv("LIBKTX_INSTALL_DIR")
LIBKTX_INCLUDE_DIR = os.getenv("LIBKTX_INCLUDE_DIR")
LIBKTX_LIB_DIR = os.getenv("LIBKTX_LIB_DIR")

if os.name == 'nt':
    if LIBKTX_INSTALL_DIR is None:
        LIBKTX_INSTALL_DIR = 'C:\\Program Files\\KTX-Software'
    if LIBKTX_INCLUDE_DIR is None:
        LIBKTX_INCLUDE_DIR = LIBKTX_INSTALL_DIR + '\\include'
    if LIBKTX_LIB_DIR is None:
        LIBKTX_LIB_DIR = LIBKTX_INSTALL_DIR + '\\lib'
elif platform.system() == 'Darwin':
    if LIBKTX_INCLUDE_DIR is None:
        LIBKTX_INCLUDE_DIR = '/usr/local/include'
    if LIBKTX_LIB_DIR is None:
        LIBKTX_LIB_DIR = '/usr/local/lib'
elif os.name == 'posix':
    if LIBKTX_INCLUDE_DIR is None:
        LIBKTX_INCLUDE_DIR = '/usr/include'
    if LIBKTX_LIB_DIR is None:
        LIBKTX_LIB_DIR = '/usr/lib'

ffibuilder = FFI()

ffibuilder.cdef(
    """
    void free(void *ptr);

    typedef struct ktxTexture ktxTexture;
    typedef struct ktxTextureCreateInfo ktxTextureCreateInfo;
    typedef struct ktxTexture1 ktxTexture1;
    typedef struct ktxHashList ktxHashList;
    typedef struct ktxHashListEntry ktxHashListEntry;

    typedef struct {
        uint32_t error;
        ktxTexture* texture;
    } ktxTextureMixed;

    typedef struct {
        void *bytes;
        size_t size;
        int error;
    } ktxWriteToMemory;

    typedef struct {
        size_t offset;
        int error;
    } ktxImageOffset;

    // Native library
    int ktxTexture_WriteToNamedFile(ktxTexture *, const char* const);
    uint32_t ktxTexture_GetElementSize(ktxTexture *);
    uint32_t ktxTexture_GetRowPitch(ktxTexture *, uint32_t level);
    size_t ktxTexture_GetImageSize(ktxTexture *, uint32_t level);
    size_t ktxTexture_GetDataSize(ktxTexture *);
    size_t ktxTexture_GetDataSizeUncompressed(ktxTexture *);
    uint8_t *ktxTexture_GetData(ktxTexture *);
    int ktxTexture_SetImageFromMemory(ktxTexture *,
                                      uint32_t level,
                                      uint32_t layer,
                                      uint32_t faceSlice,
                                      void *src,
                                      size_t srcSize);
    int ktxTexture2_TranscodeBasis(void *, int outputFormat, int transcodeFlags);
    int ktxTexture2_DeflateZstd(void *, uint32_t compressionLevel);
    uint32_t ktxTexture2_GetOETF(void *);
    bool ktxTexture2_GetPremultipliedAlpha(void *);
    bool ktxTexture2_NeedsTranscoding(void *);

    int ktxHashList_AddKVPair(ktxHashList *, const char *key, unsigned int valueLen, const void *value);
    int ktxHashList_DeleteKVPair(ktxHashList *, const char *key);
    ktxHashListEntry *ktxHashList_Next(void *entry);

    // Glue code
    ktxTextureMixed PY_ktxTexture_CreateFromNamedFile(const char* const filename,
                                                      uint32_t create_flags);
    ktxWriteToMemory PY_ktxTexture_WriteToMemory(ktxTexture *);
    ktxImageOffset PY_ktxTexture_GetImageOffset(ktxTexture *,
                                                uint32_t level,
                                                uint32_t layer,
                                                uint32_t faceSlice);
    int PY_ktxTexture_get_classId(ktxTexture *);
    bool PY_ktxTexture_get_isArray(ktxTexture *);
    bool PY_ktxTexture_get_isCompressed(ktxTexture *);
    bool PY_ktxTexture_get_isCubemap(ktxTexture *);
    bool PY_ktxTexture_get_generateMipmaps(ktxTexture *);
    uint32_t PY_ktxTexture_get_baseWidth(ktxTexture *);
    uint32_t PY_ktxTexture_get_baseHeight(ktxTexture *);
    uint32_t PY_ktxTexture_get_baseDepth(ktxTexture *);
    uint32_t PY_ktxTexture_get_numDimensions(ktxTexture *);
    uint32_t PY_ktxTexture_get_numLevels(ktxTexture *);
    uint32_t PY_ktxTexture_get_numFaces(ktxTexture *);
    uint32_t PY_ktxTexture_get_kvDataLen(ktxTexture *);
    void *PY_ktxTexture_get_kvData(ktxTexture *);
    ktxHashListEntry *PY_ktxHashList_get_listHead(ktxHashList *list);
    ktxHashList *PY_ktxTexture_get_kvDataHead(ktxTexture *);
    ktxWriteToMemory PY_ktxHashList_FindValue(ktxHashList *, const char *key);
    ktxWriteToMemory PY_ktxHashListEntry_GetKey(ktxHashListEntry *);
    ktxWriteToMemory PY_ktxHashListEntry_GetValue(ktxHashListEntry *);

    ktxTextureMixed PY_ktxTexture1_Create(uint32_t glInternalFormat,
                                          uint32_t vkFormat,
                                          uint32_t *pDfd,
                                          uint32_t baseWidth,
                                          uint32_t baseHeight,
                                          uint32_t baseDepth,
                                          uint32_t numDimensions,
                                          uint32_t numLevels,
                                          uint32_t numLayers,
                                          uint32_t numFaces,
                                          bool isArray,
                                          bool generateMipmaps,
                                          int storageAllocation);
    uint32_t PY_ktxTexture1_get_glFormat(void *);
    uint32_t PY_ktxTexture1_get_glInternalformat(void *);
    uint32_t PY_ktxTexture1_get_glBaseInternalformat(void *);
    uint32_t PY_ktxTexture1_get_glType(void *);

    ktxTextureMixed PY_ktxTexture2_Create(uint32_t glInternalFormat,
                                          uint32_t vkFormat,
                                          uint32_t *pDfd,
                                          uint32_t baseWidth,
                                          uint32_t baseHeight,
                                          uint32_t baseDepth,
                                          uint32_t numDimensions,
                                          uint32_t numLevels,
                                          uint32_t numLayers,
                                          uint32_t numFaces,
                                          bool isArray,
                                          bool generateMipmaps,
                                          int storageAllocation);
    int PY_ktxTexture2_CompressAstcEx(void *texture,
                                      bool verbose,
                                      uint32_t threadCount,
                                      uint32_t blockDimension,
                                      uint32_t mode,
                                      uint32_t quality,
                                      bool normalMap,
                                      bool perceptual,
                                      char *inputSwizzle);
    int PY_ktxTexture2_CompressBasisEx(void *texture,
                                       bool uastc,
                                       bool verbose,
                                       bool noSSE,
                                       uint32_t threadCount,
                                       uint32_t compressionLevel,
                                       uint32_t qualityLevel,
                                       uint32_t maxEndpoints,
                                       float endpointRDOThreshold,
                                       uint32_t maxSelectors,
                                       float selectorRDOThreshold,
                                       char *inputSwizzle,
                                       bool normalMap,
                                       bool separateRGToRGB_A,
                                       bool preSwizzle,
                                       bool noEndpointRDO,
                                       bool noSelectorRDO,
                                       int uastcFlags,
                                       bool uastcRDO,
                                       float uastcRDOQualityScalar,
                                       uint32_t uastcRDODictSize,
                                       float uastcRDOMaxSmoothBlockErrorScale,
                                       float uastcRDOMaxSmoothBlockStdDev,
                                       bool uastcRDODontFavorSimplerModes,
                                       bool uastcRDONoMultithreading);
    uint32_t PY_ktxTexture2_get_vkFormat(void *);
    uint32_t PY_ktxTexture2_get_supercompressionScheme(void *);
    """
)

ffibuilder.set_source(
    "pyktx.native",
    """
    #include <ktx.h>
    #include "ktx_texture.h"
    #include "ktx_texture1.h"
    #include "ktx_texture2.h"
    """,
    include_dirs=['pyktx']
                 + ([LIBKTX_INCLUDE_DIR] if LIBKTX_INCLUDE_DIR is not None else []),
    sources=['pyktx/ktx_texture.c', 'pyktx/ktx_texture1.c', 'pyktx/ktx_texture2.c'],
    libraries=['ktx'],
    library_dirs=([LIBKTX_LIB_DIR] if LIBKTX_LIB_DIR is not None else []),
    runtime_library_dirs=(([LIBKTX_LIB_DIR] if LIBKTX_LIB_DIR is not None else []) if os.name != 'nt' else None))

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)

    if 'KTX_RUN_TESTS' in os.environ and os.environ['KTX_RUN_TESTS'] == 'ON':
        suite = unittest.TestLoader().discover(os.path.join(os.path.dirname(__file__), 'tests'))
        result = unittest.TextTestRunner(verbosity=2).run(suite)

        if not result.wasSuccessful():
            sys.exit(1)
